
#include <Poco/Logger.h>
#include <Poco/Format.h>
#include <Poco/Thread.h>
#include <Poco/Net/NetException.h>
#include "Handler.hpp"


using Poco::Thread;
using Poco::Logger;
using Poco::Net::ConnectionResetException;


map<string, string> Handler::_members;

FastMutex Handler::_mutexUsers;

map<long, string> Handler::_users;

FastMutex Handler::_mutexInbox;

map<string, vector<Handler::From> > Handler::_inbox;


Handler::Handler(Message* message) : _message(message)
{
}


Handler::~Handler()
{
}


void Handler::run()
{
	Logger& log = Logger::get("server");
	
	try
	{
		while (true)
		{
			if (receive())
				break;
			process();
			if (send())
				break;
		}
	}
	catch (ConnectionResetException& exc)
	{
		log.debug("Handler::run(): connection reset");
	}
	
	delete this;
}


bool Handler::receive()
{
	bool done = false;
	Logger& log = Logger::get("server");

	if (_message->session()->status() != Session::DONE)
	{
		log.debug("Handler::receive(): reading session " + Poco::format("%ld", _message->session()->id()));
		char buf[BUF_SIZE];
		int result = _message->session()->socket()->receiveBytes(reinterpret_cast<void*>(buf), BUF_SIZE - 1);
		if (result == 0)
		{
			_message->session()->status(Session::DONE);
			log.debug("Handler::receive(): cannot read session " + Poco::format("%ld", _message->session()->id()));
		}
		else
		{
			buf[result] = '\0';
			string request = buf;
			ChatLine* line = new ChatLine;
			line->parse(request);
			_message->chatLine(line);
			log.debug("Handler::receive(): message processed");
		}
	}
	else
	{
		Message::destroy(_message);
		log.information("Handler::receive(): destroying session " + Poco::format("%ld", _message->session()->id()));
		done = true;
	}
	return done;
}


bool Handler::send()
{
	bool done = false;
	Logger& log = Logger::get("server");

	if (_message->session()->status() == Session::DONE)
	{
		log.information("Handler::send(): session " + Poco::format("%ld", _message->session()->id()) + " destroyed");
		Message::destroy(_message);
		done = true;
	}
	else
	{
		string response = _message->chatLine()->format();
		int result = _message->session()->socket()->sendBytes(reinterpret_cast<const void*>(response.c_str()), response.size());
		if (result == 0)
		{
			removeUser(_message->session()->id());
			log.information("Handler::send(): cannot send, session " + Poco::format("%ld", _message->session()->id()) + " destroyed");
			Message::destroy(_message);
		}
	}
	return done;
}


void Handler::process()
{
	Logger& log = Logger::get("server");
	log.debug("Handler::process(): processing session " + Poco::format("%ld", _message->session()->id()));
	
	if (_message->session()->status() == Session::DONE)
	{
		removeUser(_message->session()->id());
		log.debug("Handler::process(): session done, removing user");
	}
	// check commands and their proper usage
	else if (_message->chatLine()->command() == "login")
	{
		if (_message->session()->status() == Session::NEW)
		{
			string username = _message->chatLine()->username();
			string password = _message->chatLine()->password();
			if (Handler::_members.count(username) > 0 && Handler::_members[username] == password)
			{
				log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", login ok");
				_message->session()->status(Session::ACTIVE);
				_mutexUsers.lock();
				_users[_message->session()->id()] = username;
				_mutexUsers.unlock();
				_message->chatLine()->command("login ok");
			}
			else
			{
				log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", login failed");
				_message->session()->status(Session::DONE);
				removeUser(_message->session()->id());
				_message->chatLine()->command("login failed");
			}
		}
		else
		{
			log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", cannot login");
			_message->chatLine()->command("cannot login");
		}
	}
	else if (_message->chatLine()->command() == "logout")
	{
		log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", logout");
		_message->session()->status(Session::DONE);
		removeUser(_message->session()->id());
		_message->chatLine()->command("logout ok");
	}
	else if (_message->chatLine()->command() == "send")
	{
		if (_message->session()->status() == Session::ACTIVE)
		{
			if (_message->chatLine()->to() == "")
			{
				log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", send failed, no 'to' field");
				_message->chatLine()->command("send failed, no 'to' field");
			}
			else if (_message->chatLine()->chatLine() == "")
			{
				log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", send failed, no 'chat' field");
				_message->chatLine()->command("send failed, no 'chat' field");
			}
			else
			{
				_mutexUsers.lock();
				string userFrom = _users[_message->session()->id()];
				_mutexUsers.unlock();
				_mutexInbox.lock();
				_inbox[_message->chatLine()->to()].push_back(From(userFrom, _message->chatLine()->chatLine()));
				_mutexInbox.unlock();
				log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", send ok, from=" +
					Poco::format("%s", _message->chatLine()->from()) + ", chat=" + Poco::format("%s", _message->chatLine()->chatLine()));
				_message->chatLine()->command("send ok");
			}
		}
		else
		{
			log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", send failed, not authorized");
			_message->session()->status(Session::DONE);
			removeUser(_message->session()->id());
			_message->chatLine()->command("send failed, not authorized");
		}
	}
	else if (_message->chatLine()->command() == "receive")
	{
		if (_message->session()->status() == Session::ACTIVE)
		{
			_mutexUsers.lock();
			string username = _users[_message->session()->id()];
			_mutexUsers.unlock();
			_mutexInbox.lock();
			From fm;
			if (_inbox.count(username) > 0 && _inbox[username].size() > 0)
			{
				fm = _inbox[username].back();
				_inbox[username].pop_back();
			}
			_mutexInbox.unlock();
			if (fm.empty())
			{
				_message->chatLine()->command("receive ok, no chatLine");
				log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", receive ok, no chatLine");
			}
			else
			{
				_message->chatLine()->command("receive ok");
				_message->chatLine()->chatLine(fm.chatLine);
				_message->chatLine()->from(fm.from);
				log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", receive ok, from=" +
					Poco::format("%s", fm.from) + ", chat=" + Poco::format("%s", fm.chatLine));
			}
		}
		else
		{
			log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", receive failed, not authorized");
			_message->session()->status(Session::DONE);
			removeUser(_message->session()->id());
			_message->chatLine()->command("receive failed, not authorized");
		}
	}
	else
	{
		log.debug("Handler::process(): session=" + Poco::format("%ld", _message->session()->id()) + ", unknown command");
		_message->chatLine()->command("unknown command");
	}
}


/**
Removes user's session.
**/
void Handler::removeUser(long sessionId)
{
	_mutexUsers.lock();
	_users.erase(sessionId);
	_mutexUsers.unlock();
}
